function  [uroot, planinfo, plandata] = opd(xroot, model, cfg)
% OPD Optimistic planning for deterministic systems
% Arrays variant, with tree and memory reuse implemented

% Note: memory reuse will produce different results with OPD than OPSS
% implementation, even on deterministic systems. Reason: OPSS stores b_opss(x,u)
% and has the freedom to minimize b INDEPENDENTLY for each u, across all memory,
% whereas here we only store b(x) = max_u b_opss(x, u). So OPSS has more
% freedom and may produce better results. 

% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % expansion budget
CFG.criterion = 'optim';        % expansion criterion: optim(istic) or unif(orm)
CFG.U = [];                     % discrete, flat action space
% data reuse options
CFG.reusetree = 0;              % try reusing previous tree (only when k is set and > 1)
CFG.reusetree_maxsize = 0;      % do not reuse tree if larger (has more nodes) than this; 0 = unlimited size
CFG.usemem = 0;                 % use memory of b-values
CFG.lipschitz = [];             % Lipschitz constant to use in bound estimates
CFG.mem_maxsize = 0;            % 0: no limit, as large as needed, 
                                % <= 10: given as a multiple of the tree size (ignoring root)
                                % > 10: given as number of points in memory
CFG.k = -1;                     % discrete time step
% implementation options
CFG.compat = 0;                 % 0: use default, 1: use structures
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% forward call to compatibility version if configured
if cfg.compat == 1,
    [uroot, planinfo, plandata] = opd_structures(xroot, model, cfg);
    return;
end;

persistent Tparent Tchildren Tdepth Tleaf Tx Tr TR Tterm Tb treesize
persistent Mx Mb memsize
cfg = checkparams(cfg, CFG);

if cfg.reusetree && cfg.usemem, error('Memory and tree reuse cannot be used together.'); end;

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma; n = cfg.n; criterion = cfg.criterion; 
U = cfg.U; usemem = cfg.usemem; lipschitz = cfg.lipschitz;
M = size(U, 2);      % number of discrete actions
dimx = model.p;      % number of state vars

% start counting time
t = cputime;

% 1a) Memory initialization: if uninitialized, or we are just starting a new trajectory
% must be done before tree, as otherwise tree may attempt to use existing memory
if usemem && (isempty(memsize) || cfg.k <= 1),
    Mx = NaN(dimx, 0, 'double');      % states
    Mb = NaN(0, 'double');            % b-values
    memsize = 0;
end;

% 1b) Tree initialization
if cfg.reusetree,
    % an empty tree must be constructed if:
    % - no tree has ever been created
    % - or old tree too large
    % - or we are just starting a trajectory or working offline
    reusetree_clear = isempty(treesize) || ...
        (cfg.reusetree_maxsize > 0 && treesize >= cfg.reusetree_maxsize) || ... 
        cfg.k <= 1;                                 % 
end;
if ~cfg.reusetree || reusetree_clear,
    % bound on tree size, accounting for tree built at given iteration, plus maximum reused tree
    % note reusetree_maxsize=0 means "no max size", so we just take the bound w/o tree reuse
    if cfg.reusetree && cfg.reusetree_maxsize > 0,  maxtreesize = cfg.reusetree_maxsize + M*n + 1;
    else                                            maxtreesize = M*n + 1;
    end;
    % prepare tree data 
    Tparent = zeros(2, maxtreesize, 'uint32');        % parent and parent action
    Tchildren = zeros(M, maxtreesize, 'uint32');      % child index for each action
    Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
    Tleaf = false(1, maxtreesize);                    % leaf flag
    Tx = NaN(dimx, maxtreesize, 'double');            % state
    Tr = NaN(1, maxtreesize, 'double');               % reward
    TR = NaN(1, maxtreesize, 'double');               % partial return up to node (cumulative reward)
    Tterm = false(1, maxtreesize);                    % terminal flag
    Tb = NaN(1, maxtreesize, 'double');               % b-values
    
    % initialize the root node (note 0 = null pointer)
    root = 1;
    Tparent(:, root) = 0;
    Tchildren(:, root) = 0;
    Tdepth(root) = 0;
    Tleaf(root) = true;
    Tx(:, root) = xroot;
    Tr(root) = 0;
    TR(root) = 0;
    Tterm(root) = false;    % presumably, we are not asked to make a decision in a terminal state
    if usemem && memsize > 0,   % memsize >0 is false also if memsize is empty
        Tb(:, root) = estimateb(xroot, Mx, Mb, memsize, discount, lipschitz, dimx);
    else
        Tb(root) = 1/(1-discount);
    end;
    
    treesize = 1;
    if cfg.reusetree, subtreesize = 0; end;
else        % reuse tree
    % find nodes containing xroot, choose a min-depth one (larger tree to reuse)
    nodesxroot = findflat(xroot, Tx);
    if isempty(nodesxroot), error('Cannot find xroot on old tree'); end;    % shouldn't happen
    [depthmin idepthmin] = min(Tdepth(nodesxroot));
    subtreeroot = nodesxroot(idepthmin);
    % traverse subtree of this node remembering all nodes
    subtree = subtreeroot; frontier = subtreeroot;
    while ~all(Tleaf(frontier)),            % there are still inner nodes to process
        newfrontier = [];
        for node = frontier(~Tleaf(frontier)),     % add children of each inner node
            newfrontier = [newfrontier Tchildren(:, node)'];
        end;
        % the found nodes are added to the subtree, and become the new frontier
        subtree = [subtree newfrontier];
        frontier = newfrontier;
    end;
    subtreesize = length(subtree);
    
    % create the new tree
    % update node pointers
    newpointers = zeros(1, treesize+1, 'uint32'); % the extra zero is special, will be used below
    newpointers(subtree) = 1:subtreesize;
    % note root pointers to parent are 0
    Tparent(:, 1:subtreesize) = [0 newpointers(Tparent(1, subtree(2:end))); 0 Tparent(2, subtree(2:end))];
    % make the zero chidren point to newpointers(treesize+1) = 0 (so they again become zero)
    children = Tchildren(:, subtree); children = children(:); children(children == 0) = treesize + 1;
    Tchildren(:, 1:subtreesize) = reshape(newpointers(children), M, subtreesize);
    % depth decreases
    subtreedepth = double(Tdepth(subtreeroot));
    Tdepth(1:subtreesize) = Tdepth(subtree) - subtreedepth;
    Tleaf(1:subtreesize) = Tleaf(subtree);
    Tx(:, 1:subtreesize) = Tx(:, subtree);
    Tr(1:subtreesize) = Tr(subtree);
    Tterm(1:subtreesize) = Tterm(subtree);
    Tb(1:subtreesize) = Tb(subtree);
    % partial return is recalculated relative to the new root
    TR(1:subtreesize) = (TR(subtree) - TR(subtreeroot)) ./ (discount^subtreedepth);
    
    % final reassignments
    root = newpointers(subtreeroot);
    if root ~= 1, error('New root is not 1!'); end; % should never happen
    treesize = subtreesize;
    
    % these should never be used again if the code is correct, but
    % reinitialize them just in case
    Tparent(:, treesize+1:end) = 0; Tchildren(:, treesize+1:end) = 0;
    Tdepth(treesize+1:end) = 0; Tleaf(treesize+1:end) = false;
    Tx(:, treesize+1:end) = NaN; Tr(treesize+1:end) = NaN;
    Tterm(treesize+1:end) = false;
    Tb(treesize+1:end) = NaN; TR(treesize+1:end) = NaN;
end;

% 2) Main tree construction loop
for tau = 1:n,
    
    % 2a) select candidate leaves for expansion
    leaves = find(Tleaf);
    if criterion(1) == 'o', 
        % optimistic: max b-value RELATIVE TO THE ROOT
        B = TR(leaves) + discount.^double(Tdepth(leaves)) .* Tb(leaves);
        leaves = leaves(B == max(B));
        % Notes: 
        % - due to OPD structure we are able to directly go for the max-b
        % leaf, equivalent to iteratively developing all optimistic paths
        % - without memory use, Tb(leaves) will always be 1/1-gamma
    elseif criterion(1) == 'u',
        % uniform: smallest depth
        leaves = leaves(Tdepth(leaves) == min(Tdepth(leaves)));
    end;
    % select only nonterminal leaves; if none, nothing left to do
    leaves = leaves(~Tterm(leaves)); if isempty(leaves); break ;end;
    % break any ties between remaining candidates deterministically
    nodetoexpand = leaves(1);
        
    % 2b) expand the chosen node
    Tleaf(nodetoexpand) = false;        % no longer a leaf
    % run through the discrete actions 
    for j = 1:M,
        % find index where new node will be, save it on the children table of current node
        child = treesize + 1;
        Tchildren(j, nodetoexpand) = child;
        
        % simulate transition and add data to tree
        [Tx(:, child), Tr(child), Tterm(child)] = ...
            feval(model.fun, model, Tx(:, nodetoexpand), U(:, j));
        % if Tr(child) < 0 || Tr(child) > 1, error('Rewards must be in [0, 1]'); end;

        % create new node
        Tparent(:, child) = [nodetoexpand; j];      % parent is expanded node, along jth action
        Tchildren(:, child) = 0;                    % init children pointers to "null"
        Tdepth(child) = Tdepth(nodetoexpand) + 1;   % one deeper than the parent
        Tleaf(child) = true;                        % new child is a leaf
        TR(child) = TR(nodetoexpand) + discount^double(Tdepth(nodetoexpand)) * Tr(child);
        if Tterm(child),   
            Tb(child) = 0;             % values of terminal nodes exactly 0
        elseif usemem && memsize > 0,
            Tb(child) = estimateb(Tx(:, child), Mx, Mb, memsize, discount, lipschitz, dimx);
        else
            Tb(child) = 1/(1-discount);
        end;
            
        % we just added a new node, increase tree size
        treesize = treesize + 1;
    end;
    
    % 2c) update b-values affected by expanded node (path to root)
    node = nodetoexpand;
    while node ~= 0,       % when node is 0, it means the root has already been processed
        % account for the possibility of fewer successor states than the maximum L
        children = Tchildren(:, node);
        Tb(node) = max(Tr(children) + discount .* Tb(children));
        node = Tparent(1, node);
    end;

    if cfg.verb >= 4,
        visualizetree(Tparent, Tchildren, Tdepth, Tleaf, Tx, Tr, Tterm, Tb, TR, U, nodetoexpand); 
        set(gcf, 'Name', sprintf('Iter %d: Tree, optimistic path, expanded node', cfg.k));
        pause;
    end;
        
end;    % end main tree construction loop

% 3) Update memory, if using it
if usemem,
    % 3a) Aggregate duplicates (this code is based by removeduplicates.m, and customized)
    % start with existing memory and values on the new tree, w/o leaves as they are derived from
    % the existing memory so by definition they cannot provide better bounds
    innernodes = find(~Tleaf(1:treesize));
    Mxtemp = [Mx Tx(:, innernodes)]';
    Mbtemp = [Mb Tb(innernodes)]';
    memsize_initial= size(Mxtemp, 1);
    [Mxtemp, irearrange] = sortrows(Mxtemp);
    Mbtemp = Mbtemp(irearrange);
    % indices where points are different (up to numerical tolerance) -- are starting points for the index ranges
    ind = find(any(abs(diff(Mxtemp, 1)) > 1e-10, 2));
    ind = [1; 1+ind; size(Mxtemp, 1)+1];  % 1 to also consider 1st elem; size+1 for "dummy" last range
    memsize = length(ind) - 1;
    % init new memory
    Mx = zeros(dimx, memsize); Mb = zeros(1, memsize);
    % since hopefully most of the data will be unique, handle it efficiently
    indsingle = (diff(ind) == 1)';
    Mx(:, indsingle) = Mxtemp(ind([indsingle false]), :)';
    Mb(indsingle) = Mbtemp(ind([indsingle false]))';
    for i = find(~indsingle),
        % for x, take the first value -- should be fine
        Mx(:, i) = Mxtemp(ind(i), :)'; 
        % for b, choose the minimum value
        Mb(i) = min(Mbtemp(ind(i):ind(i+1)-1));
    end;
    memsize_duplicates = memsize_initial - memsize;
    
    % 3b) Remove redundant points
    Ms = zeros(1, memsize);
    for i = 1:memsize,
        bi = Mb(i);      % save b(x_i)
        Mb(i) = Inf;     % equivalent to disregarding this point when estimating bound
        % score of each point is the difference between the estimated bound /without/ the point in
        % the set, and its own bound;
        % this score is negative when the sample is redundant, i.e. its bound is larger than
        % resulting from neighbors (or than absolute bound 1/1-gamma)
        Ms(i) = estimateb(Mx(:, i), Mx, Mb, memsize, discount, lipschitz, dimx) - bi;
        Mb(i) = bi;      % restore b(x_i)
    end;
    % points that do not contribute to the aggregate bound are redundant
    keep = Ms > 0;
    % remember what we are removing if we'll be plotting it later
    if cfg.verb >= 3, Mxremoved = Mx(:, ~keep); Mbremoved = Mb(~keep);  end; 
    % shrink memory
    Mx = Mx(:, keep);
    Mb = Mb(keep);
    memsize_redundant = memsize - size(Mx, 2);
    memsize = size(Mx, 2);

    % 3c) Remove other points sorted by their score to shrink to allowed memory size
    if cfg.mem_maxsize > 0 && cfg.mem_maxsize <= 10,   
        % memory given as a fraction/multiple of tree size; exclude root to make nicer numbers
        mem_maxsize = ceil((treesize - 1) * cfg.mem_maxsize);
    else% given directly as a number
        mem_maxsize = cfg.mem_maxsize;
    end;
    if mem_maxsize > 0 && memsize > mem_maxsize,
        % the aggregate score is the maximum of scores for each action
        Ms = Ms(keep);
        [Ms, irearrange] = sort(Ms, 2, 'descend');
        % remember what we are removing if we'll be plotting it later
        if cfg.verb >= 3, 
            Mxremoved = [Mxremoved Mx(:, irearrange(mem_maxsize+1:end))]; 
            Mbremoved = [Mbremoved Mb(irearrange(mem_maxsize+1:end))]; 
        end; 
        Mx = Mx(:, irearrange(1:mem_maxsize));
        Mb = Mb(irearrange(1:mem_maxsize));
        memsize_pruned = memsize - mem_maxsize;
        memsize = size(Mx, 2);
    else
        memsize_pruned = 0;
    end;
    
    % visualize including the points that we'll erase
    if usemem && cfg.verb >= 3,
        visualizememory(Mx, Mb, cfg, Mxremoved, Mbremoved); 
        set(gcf, 'Name', sprintf('k=%d: Visualizing memory and any cleaned up points', cfg.k));
        pause; 
    end;        
end;

% 4) Find action at the tree root
% choose a leaf with maximal u-value (i.e., truncated return)
leaves = find(Tleaf); [Rmax, jmax] = max(TR(leaves)); greedy = leaves(jmax); 
% navigate greedy path upwards in the tree until level 1, just after root
while Tparent(1, greedy) ~= root, greedy = Tparent(1, greedy); end;
% choose the action that was taken at root to arrive at this level-1 node
% -- this is the action greedy in util
uroot = U(:, Tparent(2, greedy));
% Note: this avoids the need to store and backup util-values while building the tree

% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max(Tdepth);
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize); 
    if cfg.reusetree,
        planinfo.reusetree_clear = reusetree_clear;
        planinfo.reusetree_size = subtreesize;
        planinfo.message = [planinfo.message sprintf('; reusedtreesize=%d', planinfo.reusetree_size)];
    end;
    if usemem,
        planinfo = copyfields(...
            varstostruct('memsize', 'memsize_initial', 'memsize_duplicates', 'memsize_redundant', 'memsize_pruned'), ...
            planinfo);
        planinfo.message = [planinfo.message sprintf('; memsizes:init=%d/dup=%d/redund=%d/pruned=%d/final=%d', ...
            memsize_initial, memsize_duplicates, memsize_redundant, memsize_pruned, memsize)];
    end;
end;
% output only tree data (no memory)
if nargout >= 3,
    plandata.Tparent = Tparent(:, 1:treesize);
    plandata.Tchildren = Tchildren(:, 1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.TR = Tr(1:treesize);
    plandata.Tterm = Tterm(1:treesize);
    plandata.Tb = Tb(1:treesize);
end;

end % main function


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CORE LOCAL FUNCTIONS

% Estimate b-values
function b = estimateb(x, Mx, Mb, memsize, discount, lipschitz, dimx)
    % compute L1 distance to all points in memory, dimension-wise
    dist = zeros(1, memsize);
    for ii = 1:dimx, dist = dist + abs(x(ii) - Mx(ii, 1:memsize)); end;
    % find b-value estimate as minimum across points (capping at 1/1-gamma)
    b = min(min(Mb + lipschitz .* dist), 1/(1-discount));
end     % estimateb

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Other local functions

% Plot a graphical representation of the current tree
% (argument list is a mess but whatever)
function visualizetree(Tparent, Tchildren, Tdepth, Tleaf, Tx, Tr, Tterm, Tb, TR, ...
    U, expandednode)
persistent figh;
figsize = [800, 600];
if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
clf;

% blue nodes: the ones on optimistic path
node = expandednode; bluenodes = [];
while Tparent(1, node) ~= 0, node = Tparent(1, node);  bluenodes(end+1) = node; end;
% red nodes: the one just expanded and its children
rednodes = [expandednode; Tchildren(:, expandednode)];

% initialize helper variables for nested functions
M = size(U, 2);
nodestoexpand = {};
% build and customize tree
guitree = uitree('Root', rendernode(1));
guitree.Position(3) = figsize(1);   % expand to figure width
% expand nodes
for i = 1:length(nodestoexpand), guitree.expand(nodestoexpand{i}); end;
% done

    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Nested function to render a node recursively
    function guinode = rendernode(i, jp)
    % create label
    if i == 1,  label = 'root: ';
    else        label = ['u=' num2strx(U(:, jp)', 'v') ' --> '];
    end;
    label = [label sprintf('x=%s, r=%.4f, b=%.4f', num2strx(Tx(:, i)', 'v'), Tr(i), Tb(i))];
    if Tterm(i), label = [label ' T']; end;
    if any(i == bluenodes),label = ['<html><font color="blue">' label '</font></html>']; 
    elseif any(i == rednodes), label = ['<html><font color="red">' label '</font></html>']; 
    end;
    % add tree(i) to tree
    guinode = uitreenode(label, label, [], Tleaf(i));
    if ~Tleaf(i),
        % remember we need to expand this node
        nodestoexpand{end+1} = guinode;
        for jj = 1:M, guinode.add(rendernode(Tchildren(jj, i), jj)); end;
    end;
    end     % function rendernode
end     % function visualizetree



% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Plot a graphical representation of the current memory
function visualizememory(Mx, Mb, cfg, Mxremoved, Mbremoved)
dimx = size(Mx, 1);
memsize = size(Mx, 2);
discount = cfg.gamma;
lipschitz = cfg.lipschitz;

if dimx > 2, return; end;        % can only draw up to 2-D data

persistent figh
figsize = [800, 600];
if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
clf;

for i = 1:dimx, X{i} = linspace(min(Mx(i, :)), max(Mx(i, :)), 150); end;
Xflat = flat(X);
B = zeros(1, length(Xflat));
for ii = 1:length(Xflat), B(ii) = estimateb(Xflat(: , ii), Mx, Mb, memsize, discount, lipschitz, dimx); end;

if dimx == 1,
    scatter(Mx, Mb, 'filled'); hold on;
    plot(X{1}, B, 'k');
    if nargin >= 4,     % highlight removed points
        scatter(Mxremoved, Mbremoved, 'x', 'MarkerEdgeColor', 'r', 'LineWidth', 2);
    end;
    set(gca, 'xlim', minmax(Mx));
elseif dimx == 2
    surf(X{1}, X{2}, reshape(B, length(X{1}), length(X{2}))', 'EdgeColor', 'none', 'FaceAlpha', 0.5); hold on;
    scatter3(Mx(1, :), Mx(2, :), Mb, 'filled'); 
    colormap(cool);
    if nargin >= 4,     % highlight removed points
        scatter3(Mxremoved(1, :), Mxremoved(2, :), Mbremoved, 'x', 'MarkerEdgeColor', 'r', 'LineWidth', 2);
    end;
    set(gca, 'xlim', minmax(Mx(1, :)));
    set(gca, 'ylim', minmax(Mx(2, :)));
end;

end    % function visualizememory


